/*
 * Copyright (c) 1985 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "mathimpl.h"

/* HYPOT(X,Y)
 * RETURN THE SQUARE ROOT OF X^2 + Y^2  WHERE Z=X+iY
 * DOUBLE PRECISION (VAX D format 56 bits, IEEE DOUBLE 53 BITS)
 * CODED IN C BY K.C. NG, 11/28/84; 
 * REVISED BY K.C. NG, 7/12/85.
 *
 * Required system supported functions :
 *	__copysign(x,y)
 *	__finite(x)
 *	__scalb(x,N)
 *	__sqrt(x)
 *
 * Method :
 *	1. replace x by |x| and y by |y|, and swap x and
 *	   y if y > x (hence x is never smaller than y).
 *	2. Hypot(x,y) is computed by:
 *	   Case I, x/y > 2
 *		
 *				       y
 *		hypot = x + -----------------------------
 *			 		    2
 *			    sqrt ( 1 + [x/y]  )  +  x/y
 *
 *	   Case II, x/y <= 2 
 *				                   y
 *		hypot = x + --------------------------------------------------
 *				          		     2 
 *				     			[x/y]   -  2
 *			   (sqrt(2)+1) + (x-y)/y + -----------------------------
 *			 		    			  2
 *			    			  sqrt ( 1 + [x/y]  )  + sqrt(2)
 *
 *
 *
 * Special cases:
 *	hypot(x,y) is INF if x or y is +INF or -INF; else
 *	hypot(x,y) is NAN if x or y is NAN.
 *
 * Accuracy:
 * 	hypot(x,y) returns the sqrt(x^2+y^2) with error less than 1 ulps (units
 *	in the last place). See Kahan's "Interval Arithmetic Options in the
 *	Proposed IEEE Floating Point Arithmetic Standard", Interval Mathematics
 *      1980, Edited by Karl L.E. Nickel, pp 99-128. (A faster but less accurate
 *	code follows in	comments.) In a test run with 500,000 random arguments
 *	on a VAX, the maximum observed error was .959 ulps.
 *
 * Constants:
 * The hexadecimal values are the intended ones for the following constants.
 * The decimal values may be used, provided that the compiler will convert
 * from decimal to binary accurately enough to produce the hexadecimal values
 * shown.
 */

#define r2p1hi	(__math_const[MC_R2P1HI])
#define r2p1lo	(__math_const[MC_R2P1LO])
#define sqrt2	(__math_const[MC_SQRT2])

#define IBIG	30	/* fl(1+2**(2*ibig))==1 */


double
DEFUN(__hypot, (x, y), double x AND double y)
{
	double t,r;
	int exp;

	if(__finite(x))
	    if(__finite(y))
	    {	
		x=__abs(x);
		y=__abs(y);
		if(y > x) 
		    { t=x; x=y; y=t; }
		if(x == zero) return(zero);
		if(y == zero) return(x);
		exp= __logb(x);
		if(exp-(int)__logb(y) > IBIG) {
		    /* raise inexact flag and return |x| */
		    set_inexact_flag(one, small18);
		    return x;
		}

	    /* start computing sqrt(x^2 + y^2) */
		r=x-y;
		if(r>y) { 	/* x/y > 2 */
		    r=x/y;
		    r=r+__sqrt(one+r*r); }
		else {		/* 1 <= x/y <= 2 */
		    r/=y; t=r*(r+2.0);
		    r+=t/(sqrt2+__sqrt(2.0+t));
		    r+=r2p1lo; r+=r2p1hi; }

		r=y/r;
		return(x+r);

	    }

	    else if (!__isnan(y))	/* y is +-INF */
		     return __abs(y);
	    else 
		     return y;		/* y is NaN and x is finite */

	else if (!__isnan(x))		/* x is +-INF */
	         return __abs(x);
	else if(__finite(y))
	         return(x);		/* x is NaN, y is finite */
	else if(!__isnan(y)) return(y);	/* x and y is NaN */
	else return __abs(y);		/* y is INF */
}

/* CABS(Z)
 * RETURN THE ABSOLUTE VALUE OF THE COMPLEX NUMBER  Z = X + iY
 * DOUBLE PRECISION (VAX D format 56 bits, IEEE DOUBLE 53 BITS)
 * CODED IN C BY K.C. NG, 11/28/84.
 * REVISED BY K.C. NG, 7/12/85.
 *
 * Required kernel function :
 *	hypot(x,y)
 *
 * Method :
 *	cabs(z) = hypot(x,y) .
 */

double
DEFUN(__cabs, (z), struct __cabs_complex z)
{
	return __hypot(z.__x,z.__y);
}

double
DEFUN(__z_abs, (z), struct __cabs_complex *z)
{
	return __hypot(z->__x,z->__y);
}

/* A faster but less accurate version of cabs(x,y) */
#if 0
double
DEFUN(__hypot, (x,y), double x AND double y)
{
	double temp;
	int exp;

	if(__finite(x))
	    if(__finite(y))
	    {	
		x=__abs(x);
		y=__abs(y);
		if(y > x) 
		    { temp=x; x=y; y=temp; }
		if(x == zero) return(zero);
		if(y == zero) return(x);
		exp= __logb(x);
		x=__scalb(x,-exp);
		if(exp-(int)__logb(y) > IBIG) 
			/* raise inexact flag and return |x| */
		   { one+small18; return(__scalb(x,exp)); }
		else y=__scalb(y,-exp);
		return(__scalb(__sqrt(x*x+y*y),exp));
	    }

	    else if (!__isnan(y))	/* y is +-INF */
		     return(__abs(y));
	    else 
		     return(y);	   /* y is NaN and x is finite */

	else if (!__isnan(x)) 		/* x is +-INF */
	         return __abs(x);
	else if (__finite(y))
	         return(x);		/* x is NaN, y is finite */
	else if (__isnan(y)) return(y);	/* x and y is NaN */
	else return __abs(y);		/* y is INF */
}
#endif


#if (defined(__ELF__) || defined(__GNU_LIBRARY__)) && !defined(_G_NO_STABS)
#include <gnu-stabs.h>
#ifdef weak_alias
weak_alias (__cabs, cabs);
weak_alias (__z_abs, z_abs);
weak_alias (__hypot, hypot);
#endif
#endif
